@@ -6,16 +6,18 @@ import getpass
 from PySide2.QtWidgets import QApplication, QMainWindow
 from PySide2 import QtCore, QtGui, QtWidgets
 
-import famegui.qt_resources_rc
 from famegui import models
-from famegui.ui_mainwindow import Ui_MainWindow
+import famegui.generated.qt_resources_rc
+from famegui.generated.ui_mainwindow import Ui_MainWindow
+from famegui.dialog_scenario_properties import DialogScenarioProperties
 from famegui.dialog_newagent import DialogNewAgent
 from famegui.dialog_newcontract import DialogNewContract
-from famegui.dialog_scenario_properties import DialogScenarioProperties
+from famegui.dialog_select_schema import DialogSelectSchemas
 from famegui.scenario_graph_view import ScenarioGraphView
 from famegui.agent_controller import AgentController
-from famegui.scenario_controller import ScenarioController
+from famegui.maincontroller import MainController
 from famegui.appworkingdir import AppWorkingDir
+from famegui.protobuf_generation import generate_protobuf_from_scenario_yaml_file
 
 
 class PropertyTreeItem(QtWidgets.QTreeWidgetItem):
@@ -39,7 +41,7 @@ class MainWindow(QMainWindow):
         self.ui.setupUi(self)
         self._working_dir = working_dir
         self._tree_items_for_agent_types = {}
-        self._scenario_ctrl = ScenarioController(self._working_dir)
+        self._controller = MainController(self._working_dir)
         # init
         self._init_ui()
         self._connect_actions()
@@ -57,6 +59,9 @@ class MainWindow(QMainWindow):
         self.ui.graphicsView.setBackgroundBrush(QtGui.QColor(230, 230, 230))
         self.setWindowTitle(
             QtWidgets.QApplication.instance().applicationDisplayName())
+        # status bar
+        self._status_label_icon = QtWidgets.QLabel()
+        self.statusBar().addWidget(self._status_label_icon)
         # project struture tree view
         self.ui.treeProjectStructure.setSelectionMode(
             QtWidgets.QAbstractItemView.SingleSelection)
@@ -70,6 +75,8 @@ class MainWindow(QMainWindow):
         self.ui.treeAttributes.setHeaderLabels(["Attribute", "Value"])
         self.ui.treeAttributes.setColumnWidth(0, 140)
         self.ui.treeAttributes.setAlternatingRowColors(True)
+        self.ui.treeAttributes.setEditTriggers(
+            QtWidgets.QAbstractItemView.NoEditTriggers)
 
     def _connect_actions(self):
         logging.debug("conecting main window actions")
@@ -84,8 +91,13 @@ class MainWindow(QMainWindow):
         self.ui.actionSaveProjectAs.triggered.connect(self.save_project_as)
         # close
         self.ui.actionCloseProject.triggered.connect(self.close_project)
+        # generate protobuf
+        self.ui.actionMakeRunConfig.triggered.connect(self.make_run_config)
         # exit
         self.ui.actionExit.triggered.connect(self.close)
+        # edit
+        self.ui.actionScenarioProperties.triggered.connect(
+            self._on_edit_scenario_properties)
 
     def _connect_slots(self):
         logging.debug("initializing main window slots")
@@ -93,7 +105,7 @@ class MainWindow(QMainWindow):
         self.ui.sliderZoomFactor.valueChanged.connect(
             self._on_zoom_value_changed)
         self._graph_view.selected_agent_changed.connect(
-            self._scenario_ctrl.set_selected_agent_id)
+            self._controller.set_selected_agent_id)
         self._graph_view.contract_creation_requested.connect(
             self._on_contract_creation_requested)
         self._graph_view.agent_creation_requested.connect(
@@ -103,11 +115,11 @@ class MainWindow(QMainWindow):
         self.ui.lineFilterPattern.textChanged.connect(
             self._filter_pattern_changed)
         # controller
-        self._scenario_ctrl.properties.changed.connect(
-            self._on_scenario_properties_changed)
-        self._scenario_ctrl.agent_added.connect(self._on_agent_added)
-        self._scenario_ctrl.contract_added.connect(self._on_contract_added)
-        self._scenario_ctrl.selected_agent_changed.connect(
+        self._controller.project_properties.changed.connect(
+            self._on_scenario_status_changed)
+        self._controller.agent_added.connect(self._on_agent_added)
+        self._controller.contract_added.connect(self._on_contract_added)
+        self._controller.selected_agent_changed.connect(
             self._on_selected_agent_changed)
 
     def _on_zoom_value_changed(self):
@@ -119,7 +131,7 @@ class MainWindow(QMainWindow):
         self.ui.labelZoomFactor.setText("{} %".format(zoomFactor))
 
     def _on_tree_view_current_item_changed(self):
-        assert self._scenario_ctrl.has_data
+        assert self._controller.is_open
 
         selected_agent_id = None
 
@@ -128,21 +140,25 @@ class MainWindow(QMainWindow):
             # note: the given id value can be None
             selected_agent_id = tree_item.data(0, QtCore.Qt.UserRole)
 
-        self._scenario_ctrl.set_selected_agent_id(selected_agent_id)
+        self._controller.set_selected_agent_id(selected_agent_id)
 
     def _on_agent_creation_requested(self, x: int, y: int):
-        assert self._scenario_ctrl.has_data
+        assert self._controller.is_open
 
-        dlg = DialogNewAgent(self._scenario_ctrl.schema, self)
+        dlg = DialogNewAgent(self._controller.schema,
+                             self._working_dir, self)
         if dlg.exec_() != 0:
             new_agent = dlg.make_new_agent(
-                self._scenario_ctrl.generate_new_agent_id())
-            self._scenario_ctrl.add_new_agent(new_agent, x, y)
+                self._controller.generate_new_agent_id())
+            self._controller.add_new_agent(new_agent, x, y)
 
     def _on_contract_creation_requested(self, sender_id: int, receiver_id: int):
-        dlg = DialogNewContract(sender_id, receiver_id, self)
+        sender = self._controller.agent_from_id(sender_id)
+        receiver = self._controller.agent_from_id(receiver_id)
+        dlg = DialogNewContract(
+            sender, receiver, self._controller.schema, self)
         if dlg.exec_() != 0:
-            self._scenario_ctrl.add_new_contract(dlg.make_new_contract())
+            self._controller.add_new_contract(dlg.make_new_contract())
 
     def _on_selected_agent_changed(self, agent_ctrl: AgentController):
         if agent_ctrl is None:
@@ -177,7 +193,7 @@ class MainWindow(QMainWindow):
 
     def _filter_pattern_changed(self):
         pattern = self.ui.lineFilterPattern.text().lower()
-        for a in self._scenario_ctrl.agent_list:
+        for a in self._controller.agent_list:
             hide = (a.type_name.lower().find(pattern) == -1)
             a.tree_item.setHidden(hide)
 
@@ -219,7 +235,7 @@ class MainWindow(QMainWindow):
         receiver_tree_item.setIcon(0, QtGui.QIcon(u":/icons/16px-logout.png"))
 
     def _confirm_current_project_can_be_closed(self) -> bool:
-        if self._scenario_ctrl.has_unsaved_changes:
+        if self._controller.has_unsaved_changes:
             choice = QtWidgets.QMessageBox.warning(
                 self,
                 self.tr("Modifications will be lost"),
@@ -244,7 +260,7 @@ class MainWindow(QMainWindow):
         # reset attributes
         self._tree_items_for_agent_types = {}
         # reset scenario
-        self._scenario_ctrl.reset()
+        self._controller.reset()
         # reset ui
         self.ui.treeProjectStructure.clear()
         self.ui.treeAttributes.clear()
@@ -257,25 +273,24 @@ class MainWindow(QMainWindow):
         self._on_project_closed()
 
         # ask user to choose which schema to use for that new scenario
-        dlg = DialogScenarioProperties(
-            self._working_dir.list_existing_schema_names(), self)
+        dlg = DialogSelectSchemas(self._working_dir, self)
         if dlg.exec_() != 0:
-            schema_name = dlg.selected_schema_name()
-            schema = self._must_load_schema_from_name(schema_name)
-            self._scenario_ctrl.create_new_scenario(schema)
+            schema_path = dlg.selected_schema_path()
+            schema = models.SchemaLoader.load_yaml_file(schema_path)
+            self._controller.create_new_scenario(schema)
 
     def save_project(self) -> bool:
-        if not self._scenario_ctrl.has_data:
+        if not self._controller.is_open:
             return False
-        return self._do_save_project_as(self._scenario_ctrl.properties.file_path)
+        return self._do_save_project_as(self._controller.project_properties.file_path)
 
     def save_project_as(self) -> bool:
-        if not self._scenario_ctrl.has_data:
+        if not self._controller.is_open:
             return False
         return self._do_save_project_as("")
 
     def _do_save_project_as(self, file_path: str) -> bool:
-        assert self._scenario_ctrl.has_data
+        assert self._controller.is_open
 
         if file_path == "":
             file_path, _ = QtWidgets.QFileDialog.getSaveFileName(
@@ -286,7 +301,7 @@ class MainWindow(QMainWindow):
             if file_path == "":
                 return False
 
-        self._scenario_ctrl.save_to_file(file_path)
+        self._controller.save_to_file(file_path)
         return True
 
     def close_project(self) -> bool:
@@ -302,23 +317,23 @@ class MainWindow(QMainWindow):
             self,
             self.tr("Open scenario file"),
             self._working_dir.scenarios_dir,
-            "Scenario file (*.yaml *.yml)")
+            self.tr("Scenario (*.yaml *.yml)"))
         if file_path != "":
             self.load_scenario_file(file_path)
 
-    def _on_scenario_properties_changed(self):
-        has_data = self._scenario_ctrl.has_data
+    def _on_scenario_status_changed(self):
         logging.debug(
-            "scenario properties changed, has data={}".format(has_data))
+            "scenario status changed: agents={}, contracts={}".format(self._controller.agent_count, self._controller.contract_count))
 
-        self.ui.treeProjectStructure.setEnabled(has_data)
-        self.ui.treeAttributes.setEnabled(has_data)
-        self.ui.lineFilterPattern.setEnabled(has_data)
-        self.ui.graphicsView.setEnabled(has_data)
-        self.ui.sliderZoomFactor.setEnabled(has_data)
+        is_open = self._controller.is_open
+        self.ui.treeProjectStructure.setEnabled(is_open)
+        self.ui.treeAttributes.setEnabled(is_open)
+        self.ui.lineFilterPattern.setEnabled(is_open)
+        self.ui.graphicsView.setEnabled(is_open)
+        self.ui.sliderZoomFactor.setEnabled(is_open)
 
-        props = self._scenario_ctrl.properties
-        if has_data:
+        props = self._controller.project_properties
+        if is_open:
             if props.file_path != "":
                 self.ui.labelProjectName.setText(props.file_path)
             else:
@@ -327,43 +342,60 @@ class MainWindow(QMainWindow):
             self.ui.labelProjectName.setText("")
 
         self.ui.actionSaveProject.setEnabled(props.has_unsaved_changes)
-        self.ui.actionSaveProject.setEnabled(props.has_unsaved_changes)
+        self.ui.actionScenarioProperties.setEnabled(is_open)
+        self.ui.actionMakeRunConfig.setEnabled(
+            self._controller.can_export_protobuf)
+
+        # update status bar
+        if self._controller.agent_count > 0:
+            if props.is_validation_successful:
+                self._status_label_icon.setPixmap(":/icons/success-16px.png")
+                self._status_label_icon.setToolTip(
+                    self.tr("Schema validation succeeded"))
+            else:
+                self._status_label_icon.setPixmap(":/icons/warning-16px.png")
+                self._status_label_icon.setToolTip(
+                    self.tr("Schema validation failed"))
+        else:
+            self._status_label_icon.clear()
 
     def _ask_user_to_select_new_schema(self, title: str, reason_msg: str) -> str:
         QtWidgets.QMessageBox.warning(self, title, reason_msg)
-        dlg = DialogScenarioProperties(
-            self._working_dir.list_existing_schema_names(), self)
+        dlg = DialogSelectSchemas(self._working_dir, self)
         if dlg.exec_() != 0:
-            return dlg.selected_schema_name()
+            return dlg.selected_schema_path()
         return ""
 
-    def _confirm_schema_name_to_use(self, schema_name: str) -> str:
-        if schema_name == "":
-            logging.info("no schema name specified, asking user to select one")
+    def _confirm_schema_file_to_use(self, schema_path: str, scenario_path: str) -> str:
+        assert os.path.isabs(scenario_path)
+
+        if schema_path == "":
+            logging.info("no schema file specified, asking user to select one")
             return self._ask_user_to_select_new_schema(
                 self.tr("No schema defined"),
-                self.tr("This scenario does not define any schema.\n\n"
+                self.tr("This scenario does not define any schema file.\n\n"
                         "Please select an existing schema to be used for that scenario.")
             )
 
-        schema_file_path = self._working_dir.find_schema_file_path_from_name(
-            schema_name)
-        if schema_file_path is None:
+        # if not absolute, the schema path is relative to the scenario file itself, and NOT the working dir
+        original_schema_path = schema_path
+        if not os.path.isabs(schema_path):
+            scenario_dir = os.path.dirname(scenario_path)
+            schema_path = os.path.join(scenario_dir, schema_path)
+            schema_path = os.path.abspath(schema_path)
+            logging.info("schema relative file path '{}' was resolved to '{}'".format(
+                original_schema_path, schema_path))
+
+        if not os.path.isfile(schema_path):
             logging.error(
-                "failed to locate schema '{}', asking user to select a new one".format(schema_name))
+                "failed to locate schema file '{}', asking user to select a new one".format(schema_path))
             return self._ask_user_to_select_new_schema(
                 self.tr("Unknown schema"),
-                self.tr("This scenario is linked to the schema '{}' which could not be located.\n\n"
-                        "Please select an existing schema for that scenario.".format(schema_name))
+                self.tr("This scenario is linked to the schema file '{}' which could not be located.\n\n"
+                        "Please select an existing schema for that scenario.".format(original_schema_path))
             )
 
-        return schema_name
-
-    def _must_load_schema_from_name(self, schema_name: str):
-        assert schema_name != ""
-        schema_file_path = self._working_dir.find_schema_file_path_from_name(
-            schema_name)
-        return models.SchemaLoader.load_yaml_file(schema_name, schema_file_path)
+        return schema_path
 
     def load_scenario_file(self, file_path):
         self._on_project_closed()
@@ -373,23 +405,74 @@ class MainWindow(QMainWindow):
             scenario_model = models.ScenarioLoader.load_yaml_file(file_path)
 
             # make sure to attach a valid schema
-            schema_name = self._confirm_schema_name_to_use(
-                scenario_model.schema_name)
-            if schema_name == "":
+            schema_path = self._confirm_schema_file_to_use(
+                scenario_model.schema_path, file_path)
+            if schema_path == "":
                 logging.warning(
                     "no schema specified: can't finalize the load of {}".format(file_path))
                 self._on_project_closed()
                 return
 
-            logging.info("opening schema '{}'".format(schema_name))
-            schema = self._must_load_schema_from_name(schema_name)
+            logging.info("opening schema file '{}'".format(schema_path))
+            schema = models.SchemaLoader.load_yaml_file(schema_path)
 
-            self._scenario_ctrl.open_scenario(
+            self._controller.open_scenario(
                 scenario_model, file_path, schema)
         except:
             self._on_project_closed()
             raise
 
+        props = self._controller.project_properties
+        if not props.is_validation_successful:
+            QtWidgets.QMessageBox.warning(self,
+                                          self.tr("Validation failure"),
+                                          self.tr("The loaded scenario does not fulfill the schema:\n\n") +
+                                          "\n".join(props.validation_errors))
+
+    def _on_edit_scenario_properties(self):
+        dlg = DialogScenarioProperties(
+            self._controller.scenario.properties, self._working_dir, self)
+        dlg.setWindowTitle(self.tr("Scenario properties"))
+        if dlg.exec_() != 0:
+            self._controller.set_scenario_properties(dlg.make_properties())
+
+    def make_run_config(self):
+        assert self._controller.can_export_protobuf
+        scenario_name = os.path.basename(
+            self._controller.project_properties.file_path).replace(".yaml", "")
+        output_path = "{}/{}.pb".format(
+            self._working_dir.protobuf_dir, scenario_name)
+        output_path = self._working_dir.make_relative_path(output_path)
+
+        dlg = DialogScenarioProperties(
+            self._controller.scenario.properties, self._working_dir, self)
+        dlg.setWindowTitle(self.tr("Make run config"))
+        dlg.enable_outputfile_selection(output_path)
+        if dlg.exec_() != 0:
+            self._controller.set_scenario_properties(dlg.make_properties())
+            self.save_project()
+            output_path = self._working_dir.make_full_path(
+                dlg.get_output_file_path())
+
+            progressDlg = QtWidgets.QProgressDialog(self)
+            progressDlg.setLabelText(self.tr("Generating protobuf file..."))
+            progressDlg.setRange(0, 0)
+            progressDlg.setCancelButton(None)
+
+            try:
+                progressDlg.show()
+                QApplication.processEvents()
+                generate_protobuf_from_scenario_yaml_file(self._working_dir.root_dir,
+                                                          self._controller.project_properties.file_path,
+                                                          output_path)
+            finally:
+                progressDlg.close()
+
+            QtWidgets.QMessageBox.information(
+                self,
+                self.tr("Operation succeeded"),
+                self.tr("The following file was successfully generated:\n{}").format(output_path))
+
     # prevent data loss when closing the main window
     def closeEvent(self, event):
         if not self._confirm_current_project_can_be_closed():
