@@ -1,4 +1,5 @@
 import typing
+import os
 
 from famegui import models
 
@@ -6,14 +7,14 @@ from famegui import models
 class Schema:
     """ Represents a schema to be applied on a scenario """
 
-    def __init__(self, name: str):
-        assert name != ""
-        self._name = name
+    def __init__(self, file_path: str):
+        assert os.path.isabs(file_path)
+        self._file_path = file_path
         self._agent_types = {}
 
     @property
-    def name(self) -> str:
-        return self._name
+    def file_path(self) -> str:
+        return self._file_path
 
     @property
     def agent_types(self) -> typing.Dict[str, models.SchemaAgentType]:
